import UIKit

// You should nowknow what a protocol is. Now, we need to talk about
// delegation.
//
//
// DELEGATION is a very commonly used DESIGN PATTERN in iOS development.
// A DESIGN PATTERN is just a way to organize your code (so as to be conducive
// to "good software design"). The delegation pattern is used extensively
// throughout iOS libraries. 
//
// Delegation is when an object, instead of performing some task, DELEGATES
// it to so other class. The "other class" is known as the DELEGATE. Often
// times, in order to define what tasks a delegate can perform, we'll use
// a protocol.
//
// Here's a very basic example of the delegation pattern:
//

protocol HelloWorldDelegate {
    func helloWorldString() -> String
}

class HelloWorldHelper : HelloWorldDelegate {
    func helloWorldString() -> String {
        return "Hello, world!"
    }
}

class HelloWorld {
    var delegate : HelloWorldDelegate?
    
    func helloWorld() {
        if (self.delegate != nil) {
            println(self.delegate!.helloWorldString())
        }
    }
}

let helloWorldObject: HelloWorld = HelloWorld()
helloWorldObject.helloWorld() // Nothing prints.
let helloWorldHelper: HelloWorldHelper = HelloWorldHelper()
helloWorldObject.delegate = helloWorldHelper
helloWorldObject.helloWorld() // Now something prints.

// So, by itself, the HelloWorld class could not complete the helloWorld()
// function. However, once we gave helloWorldObject a delegate, the function
// call worked!
//
// So, HelloWorld delegates some work to HelloWorldHelper. We use a protocol
// though because there might be multiple ways to perform helloWorldString.
// For example:

class HolaMundoHelper : HelloWorldDelegate {
    func helloWorldString() -> String {
        return "¡Hola, Mundo!"
    }
}

let holaMundoHelper: HolaMundoHelper = HolaMundoHelper()
helloWorldObject.delegate = holaMundoHelper
helloWorldObject.helloWorld() // Prints "¡Hola, Mundo!"

// So, we actually *changed the functionality* of HelloWorld without changing
// any functions in HelloWorld. This is because we basically hot swapped
// the delegate which produced the HelloWorld string with another compatible
// delegate. Now, our HelloWorld program supports internationalization. :)

// If you need another concrete example, let's play pretend for a little bit.
// Suppose the iPhone 9 comes with a brand new, very experimental piece of
// hardware called the Morbulator. The Morbulator figures out how many
// Morbulons are in the air around your iPhone (Morbulons are newly discovered
// particles which can be used as alternative energy--particles physics got
// real weird).
//
// So you work for Apple now and you need to write a class so that people can
// use your Morbulator. Let's call it the MorbulatorManager. Your boss says
// the MorbulatorManager class should be implemented as follows:
//
// 1. There should be a startTrackingMorbulons() method which turns on the
//    Morbulator hardware. It will retrieve a new reading of Morbulons every
//    second.
// 2. There should be a stopTrackingMorbulons() method which turns off the
//    Morbulator hardware.
// 3. Users of the MorbulatorManager class should define a delegate to receive
//    the raw results of the Morbulator. New results will be available every
//    second.
//
// So, your boss wants the users of the MorbulatorManager class to define a
// delegate so that users can receive results from the Morbulator. It's an
// interesting design pattern, but can you think of why this would be the case?
// It's apparently "good design."
//
// So, first we should define a protocol for interacting with the
// MorbulatorManager class. Remember, this protocol is meant to define a
// delegate to the MorbulatorManager, and so will be implemented by a different
// class, and will be called by the MorbulatorManager. This is how, in effect,
// the MorbulatorManager will "send" results to the user.

    
@objc protocol MorbulatorManagerDelegate {
    // manager: The MorbulatorManager that is passing in the readings.
    // data: The number of Morbulons within 1 cubic foot of the iPhone.
    optional func didReceiveMorbulonReading(manager: MorbulatorManager,
                                            data: UInt)
}

// The above protocol will define how a class will receive data from the
// MorbulatorManager. The client class will implement this function and
// will be passed in a reference to the MorbulatorManager that made the
// reading, as well as the data. The client class will also set itself as
// the data Then, the MorbulatorManager will call this
// method on
//
// Now we define the class. We must prepend it with @objc because we're using
// it in an @objc protocol. Don't worry about knowing these things--Xcode will
// happily telling you exactly what you're doing wrong and where.
@objc class MorbulatorManager {
    var delegate: MorbulatorManagerDelegate?
    
    func startTrackingMorbulons() {
        // start a timer to call retrieveResults().
        println("Tracking morbulons...")
    }
    
    func stopTrackingMorbulons() {
        // stop the timer to call retrieveResults()
        println("No longer tracking morbulons...")
    }
    
    // this function is called every second while we're tracking Morbulons.
    private func retrieveResults() {
        var results: UInt = 0
        // some scary ASM to get results
        
        // now we pass the results to the delegate.
        if (self.delegate != nil && self.delegate?.didReceiveMorbulonReading != nil) {
            self.delegate!.didReceiveMorbulonReading!(self, data: results)
        }
    }
}

// So this might seem a weird way to do thing. But on the user side, we can
// simply say:

class MyViewController : MorbulatorManagerDelegate {
    let morbulatorManager = MorbulatorManager()
    
    func viewDidLoad() {
        // ... some viewDidLoad code
        
        // Must set self as delegate!
        self.morbulatorManager.delegate = self
    }
    
    // Client's implementation of what to do with the Morbulon reading.
    func didReceiveMorbulonReading(manager: MorbulatorManager, data: UInt) {
        manager.stopTrackingMorbulons()
        println(data)
    }
}

// So, in this example MyViewController has a function, didReceiveMorbulonReading,
// which is called by the MorbulonManager every time a reading is made. The
// above implementation shuts off the Morbulon reading and prints the data to
// the screen.

// Of course, there can be different implementations of what to do with the
// Morbulon reading. The ViewController at this point could pass the delegation
// to a different class which might, for example, write the readings to a file.
// So, this design pattern is hot-swappable. Nice.

// This is an extremely common design pattern in iOS, so you should get used
// to it. :)
